/***************************************************************************
                          decoder_thread.cpp  -  description
                             -------------------
    begin                : Thu Feb 23 2009
    copyright            : (C) 2009 by Martin Witte
    email                : emw-kradio@nocabal.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <QtGui/QApplication>

/*#include "soundstream_decoding_step_event.h"*/
#include "soundstream_decoding_terminate_event.h"
#include "decoder_thread.h"

DecoderThread::DecoderThread(QObject *parent, const InternetRadioStation &rs, int max_buffers)
:   m_decoderOpened (false),
    m_av_pFormatCtx (NULL),
    m_av_audioStream(-1),
    m_av_aCodecCtx  (NULL),
    m_av_aCodec     (NULL),

    m_parent        (parent),
    m_RadioStation  (rs),

    m_error         (false),
    m_errorString   (QString()),
    m_done          (false),

    m_decodedSize   (0),
    m_inputURL      (m_RadioStation.url()),
    m_bufferAccessLock(1),
    m_bufferCountSemaphore(max_buffers)
{
}

DecoderThread::~DecoderThread()
{
    flushBuffers();
    closeAVStream();
}

void DecoderThread::run()
{
    openAVStream(m_inputURL.pathOrUrl());

    AVPacket pkt;

    time_t  start_time      = time(NULL);

    while (!m_error && !m_done && m_decoderOpened && (av_read_frame(m_av_pFormatCtx, &pkt) >= 0)) {

        if (!m_done && pkt.stream_index == m_av_audioStream) {

            uint8_t *audio_pkt_data = pkt.data;
            int      audio_pkt_size = pkt.size;

            while (!m_error && !m_done && m_decoderOpened && (audio_pkt_size > 0)) {
                char output_buf[(AVCODEC_MAX_AUDIO_FRAME_SIZE * 3) / 2];
                int  generated_output_bytes = sizeof(output_buf);
                int  processed_input_bytes  = avcodec_decode_audio2(m_av_aCodecCtx,
                                                                    (int16_t *)output_buf,
                                                                    &generated_output_bytes,
                                                                    audio_pkt_data,
                                                                    audio_pkt_size);
                if (processed_input_bytes < 0) {
                    /* if error, skip frame */
                    addErrorString(i18n("%1: error decoding packet. Discarding packet\n").arg(m_inputURL.pathOrUrl()));
                    break;
                }
                else if (generated_output_bytes > 0) {
                    time_t cur_time = time(NULL);
                    SoundMetaData  md(m_decodedSize, cur_time - start_time, cur_time, i18n("internal stream, not stored (%1)", m_inputURL.pathOrUrl()));

                    DataBuffer buf(output_buf, generated_output_bytes, md, m_soundFormat);

//                     printf ("free buffers: %i\n", m_bufferCountSemaphore.available());
//                     printf ("storing decoded data ...");
                    pushBuffer(buf);
//                     printf (" done\n");

/*                    printf ("semaphore available: %i\n", m_runSemaphore.available());
                    m_runSemaphore.acquire(1);
                    if (!m_done) {
                        SoundStreamDecodingStepEvent *e = new SoundStreamDecodingStepEvent(m_soundFormat, output_buf, generated_output_bytes, md);
                        QApplication::postEvent(m_parent, e);
                    }*/
                    m_decodedSize += generated_output_bytes;
                }

                audio_pkt_data += processed_input_bytes;
                audio_pkt_size -= processed_input_bytes;

            }
        }

        av_free_packet(&pkt);
//         printf ("waiting for next packet\n");
    }
//     printf ("posting event\n");
    QApplication::postEvent(m_parent, new SoundStreamEncodingTerminatedEvent(m_soundFormat));

//     printf ("closing stream\n");
    closeAVStream();
    exit();
}


void DecoderThread::pushBuffer(const DataBuffer &buf)
{
    m_bufferCountSemaphore.acquire();
    m_bufferAccessLock.acquire();
    m_buffers.push_back(buf);
    m_bufferAccessLock.release();
}


int DecoderThread::availableBuffers()
{
    int n = 0;
    m_bufferAccessLock.acquire();
    n = m_buffers.size();
    m_bufferAccessLock.release();
    return n;
}

DataBuffer &DecoderThread::getFirstBuffer()
{
    DataBuffer *tmp;
    m_bufferAccessLock.acquire();
    tmp = &(m_buffers.first());
    m_bufferAccessLock.release();
    return *tmp;
}

void DecoderThread::popFirstBuffer()
{
    m_bufferAccessLock.acquire();
    DataBuffer &buf = m_buffers.first();
    buf.freeData();
    m_buffers.pop_front();
    m_bufferAccessLock.release();
    m_bufferCountSemaphore.release();
}

void DecoderThread::flushBuffers()
{
    m_bufferAccessLock.acquire();
    while (m_buffers.size()) {
        DataBuffer &buf = m_buffers.first();
        buf.freeData();
        m_buffers.pop_front();
        m_bufferCountSemaphore.release();
    }
    m_bufferAccessLock.release();
}

void DecoderThread::setDone()
{
    m_done = true;
}

void DecoderThread::addErrorString(const QString &s)
{
    m_error = true;
    if (m_errorString.length()) {
        m_errorString += " \n";
    }
    m_errorString += s;
}


void DecoderThread::openAVStream(const QString &stream)
{
    if (m_decoderOpened) {
        return;
    }

//     fprintf (stderr, "starting up mpeg decoder test\n");

    av_register_all();

    AVInputFormat   *iformat = av_find_input_format(m_RadioStation.decoderClass().toLocal8Bit());

//     fprintf (stderr, "opening %s\n", argv[1]);

    if (av_open_input_file(&m_av_pFormatCtx, stream.toUtf8(), iformat, 0, NULL) != 0) {
        addErrorString(i18n("Could not open Stream %1").arg(stream));
        return; // Couldn't open file
    }
//     fprintf (stderr, "%s opened\n", argv[1]);

//     fprintf (stderr, "looking for stream info\n");
    // Retrieve stream information
    if (av_find_stream_info(m_av_pFormatCtx) < 0) {
        addErrorString(i18n("Could not find stream information in %1").arg(stream));
        return; // Couldn't find stream information
    }
//     fprintf (stderr, "found stream info\n");

    // Dump information about file onto standard error
//     fprintf (stderr, "dumping format\n");
//     dump_format(pFormatCtx, 0, argv[1], 0);
//     fprintf (stderr, "dumping format done\n");


    // Find the first video stream
    m_av_audioStream = -1;
//     fprintf (stderr, "looking for audio stream\n");
    for (unsigned int i = 0; i < m_av_pFormatCtx->nb_streams; i++) {
        if (m_av_pFormatCtx->streams[i]->codec->codec_type == CODEC_TYPE_AUDIO && m_av_audioStream < 0) {
            m_av_audioStream = i;
//             fprintf (stderr, "found audio stream, index %i\n", audioStream);
        }
    }

    if (m_av_audioStream == -1) {
        addErrorString(i18n("Could not find an audio stream in %1").arg(stream));
        return;
    }
//     fprintf (stderr, "found audio stream, index %i\n", audioStream);


    m_av_aCodecCtx = m_av_pFormatCtx->streams[m_av_audioStream]->codec;


    m_soundFormat = SoundFormat(m_av_aCodecCtx->sample_rate, m_av_aCodecCtx->channels, 16, true);


//     fprintf (stderr, "searching decoder\n");

    m_av_aCodec = avcodec_find_decoder(m_av_aCodecCtx->codec_id);
    if (!m_av_aCodec) {
        addErrorString (i18n("Could not find a codec for %1").arg(stream));
        return;
    }
//     fprintf (stderr, "found decoder\n");


    if (avcodec_open(m_av_aCodecCtx, m_av_aCodec) < 0) {
        addErrorString(i18n("Opening codec for %1 failed").arg(stream));
    }

    m_decoderOpened = true;
}


void DecoderThread::closeAVStream()
{
    if (!m_decoderOpened) {
        return;
    }

    avcodec_close      (m_av_aCodecCtx);
    av_close_input_file(m_av_pFormatCtx);

    m_av_pFormatCtx  = NULL;
    m_av_audioStream = -1;
    m_av_aCodecCtx   = NULL;
    m_av_aCodec      = NULL;
    m_decoderOpened  = false;
}


